use std::{
	io::{self, Write},
	path::PathBuf,
	str::FromStr,
};

use rusqlite::{Connection, params_from_iter};

use crate::{
	DB_PATH,
	cli::{self, FilesArgs, TagArgs, TagsArgs},
	try_database,
};

pub fn handle_files(args: FilesArgs) {
	try_database();

	let conn = Connection::open(DB_PATH).unwrap();
	let mut sql = r#"SELECT path, group_concat(name) as tags FROM file_tag
			JOIN file ON file.id = file_tag.file_id
			JOIN tag ON tag.id = file_tag.tag_id"#
		.to_string();

	if let Some(query) = args.query {
		let path = PathBuf::from_str(&query).unwrap_or_default();
		match path.exists() {
			true => {
				sql.push_str(&format!(" WHERE path IN ('{}')", query));
			}
			false => todo!(),
		};
	}

	sql.push_str(" GROUP BY path");

	let mut stmt = conn.prepare(&sql).unwrap();
	let result = stmt
		.query_map([], |row| -> Result<(String, String), rusqlite::Error> {
			Ok((row.get(0).unwrap(), row.get(1).unwrap()))
		})
		.unwrap()
		.flatten();

	let mut w = io::stdout();
	for (file, tags) in result {
		writeln!(&mut w, "{}: {}", file, tags).unwrap();
	}
	w.flush().unwrap();
}

pub fn handle_tag(args: TagArgs) {
	try_database();

	let mut conn = Connection::open(DB_PATH).unwrap();

	tag_file(&mut conn, args.file, args.tags);
}

pub fn handle_tags(args: TagsArgs) {
	try_database();

	let conn = Connection::open(DB_PATH).unwrap();

	match args.commands {
		Some(cli::TagsCommands::List) | None => {
			let mut w = io::stdout();
			let tags = list_tags(&conn);
			for tag in tags {
				writeln!(&mut w, "{}", tag).unwrap();
			}
			w.flush().unwrap();
		}
		Some(cli::TagsCommands::Add { add }) => add_tags(&conn, add),
		Some(cli::TagsCommands::Remove { remove }) => remove_tags(&conn, remove),
	};
}

fn list_tags(conn: &Connection) -> Vec<String> {
	let mut stmt = conn.prepare("SELECT name FROM tag").unwrap();
	let result = stmt.query_map([], |row| row.get(0)).unwrap();

	let mut tags = Vec::new();
	for name in result {
		tags.push(name.unwrap());
	}

	tags
}

fn remove_tags(conn: &Connection, tags: Vec<String>) {
	let mut query = r#"DELETE FROM tag WHERE name IN ("#.to_string();

	for (i, _tag) in tags.iter().enumerate() {
		query.push('?');
		if i < tags.len() - 1 {
			query.push(',');
		}
	}

	query.push(')');

	conn.execute(&query, params_from_iter(tags)).unwrap();
}

fn add_tags(conn: &Connection, tags: Vec<String>) {
	let mut query = r#"INSERT INTO tag(name) VALUES"#.to_string();

	for (i, _tag) in tags.iter().enumerate() {
		query.push_str("(?)");
		if i < tags.len() - 1 {
			query.push(',');
		}
	}

	conn.execute(&query, params_from_iter(tags)).unwrap();
}

fn tag_file(conn: &mut Connection, file: PathBuf, tags: Vec<String>) {
	let file = file.to_str().unwrap().to_string();

	let tx = conn.transaction().unwrap();
	{
		let file_id: Option<i32> = tx
			.query_row(r#"SELECT id FROM file WHERE path = ?"#, [&file], |row| {
				row.get(0)
			})
			.ok();

		let file_id = match file_id {
			Some(id) => id,
			None => {
				let mut stmt = tx
					.prepare(r#"INSERT INTO file(path) VALUES (?) RETURNING id"#)
					.unwrap();
				stmt.query_row([file], |row| -> Result<i32, rusqlite::Error> { row.get(0) })
					.unwrap()
			}
		};

		let mut sql = "SELECT id FROM tag WHERE name IN ".to_string();
		sql.push('(');
		let size = tags.len();
		for (i, _tag) in tags.iter().enumerate() {
			sql.push('?');
			if i < size - 1 {
				sql.push(',');
			}
		}
		sql.push(')');

		let mut stmt = tx.prepare(&sql).unwrap();
		let result = stmt
			.query_map(
				params_from_iter(tags),
				|row| -> Result<i32, rusqlite::Error> { row.get(0) },
			)
			.unwrap();

		let mut tags_ids = Vec::new();
		for id in result.flatten() {
			tags_ids.push(id);
		}

		let mut stmt = tx
			.prepare(r#"INSERT INTO file_tag(file_id, tag_id) VALUES (?, ?)"#)
			.unwrap();

		for tag_id in tags_ids {
			stmt.execute((file_id, tag_id)).unwrap();
		}
	}
	tx.commit().unwrap();
}
